# standard libraries
import contextlib
import datetime
import json
import logging
import os
import pathlib
import unittest
import uuid

# third party libraries
import numpy

# local libraries
from nion.data import Calibration
from nion.data import DataAndMetadata
from nion.swift.model import DataItem
from nion.swift.model import ImportExportManager
from nion.swift.model import Utility
from nion.swift.test import TestContext
from nion.utils import DateTime


class TestImportExportManagerClass(unittest.TestCase):

    def setUp(self):
        TestContext.begin_leaks()

    def tearDown(self):
        TestContext.end_leaks(self)

    def test_convert_data_element_records_time_zone_in_data_item_metadata(self):
        data_element = dict()
        data_element["version"] = 1
        data_element["data"] = numpy.zeros((16, 16), dtype=numpy.double)
        data_element["datetime_modified"] = Utility.get_current_datetime_item()
        with contextlib.closing(ImportExportManager.create_data_item_from_data_element(data_element)) as data_item:
            self.assertIsNotNone(data_item.created)
            self.assertEqual(data_item.timezone_offset, data_element["datetime_modified"]["tz"])

    def test_convert_data_element_sets_timezone_and_timezone_offset_if_present(self):
        data_element = dict()
        data_element["version"] = 1
        data_element["data"] = numpy.zeros((16, 16), dtype=numpy.double)
        data_element["datetime_modified"] = {'tz': '+0300', 'dst': '+60', 'local_datetime': '2015-06-10T19:31:52.780511', 'timezone': 'Europe/Athens'}
        with contextlib.closing(ImportExportManager.create_data_item_from_data_element(data_element)) as data_item:
            self.assertIsNotNone(data_item.created)
            self.assertEqual(data_item.timezone, "Europe/Athens")
            self.assertEqual(data_item.timezone_offset, "+0300")

    def test_date_formats(self):
        data_item = DataItem.DataItem()
        with contextlib.closing(data_item):
            data_item.created = datetime.datetime(2013, 11, 18, 14, 5, 4, 0)
            self.assertIsNotNone(data_item.created_local_as_string)

    def test_sub_area_size_change(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_element = dict()
            data_element["version"] = 1
            data_element["data"] = numpy.zeros((16, 16), dtype=numpy.double)
            data_item = ImportExportManager.create_data_item_from_data_element(data_element)
            document_model.append_data_item(data_item)
            self.assertEqual(data_item.dimensional_shape, (16, 16))
            self.assertEqual(data_item.data_dtype, numpy.double)
            data_element["data"] = numpy.zeros((8, 8), dtype=numpy.double)
            data_element["sub_area"] = ((0,0), (4, 8))
            ImportExportManager.update_data_item_from_data_element(data_item, data_element)
            self.assertEqual(data_item.dimensional_shape, (8, 8))
            self.assertEqual(data_item.data_dtype, numpy.double)
            data_element["data"] = numpy.zeros((8, 8), dtype=float)
            data_element["sub_area"] = ((0,0), (4, 8))
            ImportExportManager.update_data_item_from_data_element(data_item, data_element)
            self.assertEqual(data_item.dimensional_shape, (8, 8))
            self.assertEqual(data_item.data_dtype, float)

    def test_ndata_write_to_then_read_from_temp_file(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            current_working_directory = os.getcwd()
            file_path = os.path.join(current_working_directory, "__file.ndata")
            handler = ImportExportManager.NDataImportExportHandler("ndata1-io-handler", "ndata", ["ndata"])
            data_item = DataItem.DataItem(numpy.zeros((16, 16), dtype=numpy.double))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            handler.write_display_item(display_item, pathlib.Path(file_path), "ndata")
            self.assertTrue(os.path.exists(file_path))
            try:
                data_items = handler.read_data_items("ndata", pathlib.Path(file_path))
                self.assertEqual(len(data_items), 1)
                data_item = data_items[0]
                for data_item in data_items:
                    data_item.close()
            finally:
                os.remove(file_path)

    def test_npy_write_to_then_read_from_temp_file(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            current_working_directory = os.getcwd()
            file_path_npy = os.path.join(current_working_directory, "__file.npy")
            file_path_json = os.path.join(current_working_directory, "__file.json")
            numpy.save(file_path_npy, numpy.zeros((16, 16)))
            with open(file_path_json, "w") as f:
                json.dump({"version": 1}, f)
            handler = ImportExportManager.NumPyImportExportHandler("numpy-io-handler", "npy", ["npy"])
            try:
                data_items = handler.read_data_items("npy", pathlib.Path(file_path_npy))
                self.assertEqual(len(data_items), 1)
                data_item = data_items[0]
                for data_item in data_items:
                    data_item.close()
            finally:
                os.remove(file_path_npy)
                os.remove(file_path_json)

    def test_standard_io_write_to_then_read_from_temp_file(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            current_working_directory = os.getcwd()
            extensions = ("jpeg", "png", "gif", "bmp", "webp")
            for extension in extensions:
                file_path = os.path.join(current_working_directory, f"__file.{extension}")
                handler = ImportExportManager.StandardImportExportHandler(f"{extension}-io-handler", extension, [extension])
                data_item = DataItem.DataItem(numpy.zeros((16, 16), dtype=numpy.double))
                document_model.append_data_item(data_item)
                display_item = document_model.get_display_item_for_data_item(data_item)
                handler.write_display_item(display_item, pathlib.Path(file_path), extension)
                self.assertTrue(os.path.exists(file_path))
                try:
                    data_items = handler.read_data_items(extension, pathlib.Path(file_path))
                    self.assertEqual(len(data_items), 1)
                    for data_item in data_items:
                        data_item.close()
                finally:
                    os.remove(file_path)

    def test_get_writers_for_empty_data_item_returns_valid_list(self):
        data_item = DataItem.DataItem()
        with contextlib.closing(data_item):
            writers = ImportExportManager.ImportExportManager().get_writers_for_data_item(data_item)
            self.assertEqual(len(writers), 0)

    def test_get_writers_for_float_2d_data_item_returns_valid_list(self):
        data_item = DataItem.DataItem(numpy.zeros((8, 8), float))
        with contextlib.closing(data_item):
            writers = ImportExportManager.ImportExportManager().get_writers_for_data_item(data_item)
            self.assertTrue(len(writers) > 0)

    def test_data_element_date_gets_set_as_data_item_created_date(self):
        data_element = dict()
        data_element["version"] = 1
        data_element["data"] = numpy.zeros((16, 16), dtype=numpy.double)
        data_element["datetime_modified"] = {'tz': '+0300', 'dst': '+60', 'local_datetime': '2015-06-10T19:31:52.780511'}
        with contextlib.closing(ImportExportManager.create_data_item_from_data_element(data_element)) as data_item:
            self.assertIsNotNone(data_item.created)
            self.assertEqual(data_item.timezone_offset, "+0300")
            local_offset_seconds = int(round((datetime.datetime.now() - DateTime.utcnow()).total_seconds()))
            # check both matches for DST
            match1 = datetime.datetime(year=2015, month=6, day=10, hour=19 - 3, minute=31, second=52, microsecond=780511) + datetime.timedelta(seconds=local_offset_seconds)
            match2 = datetime.datetime(year=2015, month=6, day=10, hour=19 - 3, minute=31, second=52, microsecond=780511) + datetime.timedelta(seconds=local_offset_seconds + 3600)
            self.assertTrue(data_item.created_local == match1 or data_item.created_local == match2)

    def test_data_element_with_uuid_assigns_uuid_to_data_item(self):
        data_element = dict()
        data_element["version"] = 1
        data_element["data"] = numpy.zeros((16, 16), dtype=numpy.double)
        data_element_uuid = uuid.uuid4()
        data_element["uuid"] = str(data_element_uuid)
        with contextlib.closing(ImportExportManager.create_data_item_from_data_element(data_element)) as data_item:
            self.assertEqual(data_item.uuid, data_element_uuid)

    def test_creating_data_element_with_sequence_data_makes_correct_data_item(self):
        data_element = dict()
        data_element["version"] = 1
        data_element["data"] = numpy.zeros((4, 16, 16), dtype=numpy.double)
        data_element["is_sequence"] = True
        data_element["collection_dimension_count"] = 0
        data_element["datum_dimension_count"] = 2
        with contextlib.closing(ImportExportManager.create_data_item_from_data_element(data_element)) as data_item:
            self.assertEqual(data_item.is_sequence, True)
            self.assertEqual(data_item.collection_dimension_count, 0)
            self.assertEqual(data_item.datum_dimension_count, 2)
            self.assertEqual(data_item.xdata.is_sequence, True)
            self.assertEqual(data_item.xdata.collection_dimension_count, 0)
            self.assertEqual(data_item.xdata.datum_dimension_count, 2)

    def test_creating_data_element_with_sequence_and_implicit_datum_size_data_makes_correct_data_item(self):
        data_element = dict()
        data_element["version"] = 1
        data_element["data"] = numpy.zeros((4, 16, 16), dtype=numpy.double)
        data_element["is_sequence"] = True
        with contextlib.closing(ImportExportManager.create_data_item_from_data_element(data_element)) as data_item:
            self.assertEqual(data_item.is_sequence, True)
            self.assertEqual(data_item.collection_dimension_count, 0)
            self.assertEqual(data_item.datum_dimension_count, 2)
            self.assertEqual(data_item.xdata.is_sequence, True)
            self.assertEqual(data_item.xdata.collection_dimension_count, 0)
            self.assertEqual(data_item.xdata.datum_dimension_count, 2)

    def test_data_element_to_extended_data_conversion(self):
        data = numpy.ones((8, 6), int)
        intensity_calibration = Calibration.Calibration(offset=1, scale=1.1, units="one")
        dimensional_calibrations = [Calibration.Calibration(offset=2, scale=2.1, units="two"), Calibration.Calibration(offset=3, scale=2.2, units="two")]
        metadata = {"hardware_source": {"one": 1, "two": "b"}}
        timestamp = datetime.datetime.now()
        data_descriptor = DataAndMetadata.DataDescriptor(is_sequence=False, collection_dimension_count=1, datum_dimension_count=1)
        xdata = DataAndMetadata.new_data_and_metadata(data, intensity_calibration=intensity_calibration, dimensional_calibrations=dimensional_calibrations, metadata=metadata, timestamp=timestamp, data_descriptor=data_descriptor)
        data_element = ImportExportManager.create_data_element_from_extended_data(xdata)
        new_xdata = ImportExportManager.convert_data_element_to_data_and_metadata(data_element)
        self.assertTrue(numpy.array_equal(data, new_xdata.data))
        self.assertNotEqual(id(new_xdata.intensity_calibration), id(intensity_calibration))
        self.assertEqual(new_xdata.intensity_calibration, intensity_calibration)
        self.assertNotEqual(id(new_xdata.dimensional_calibrations[0]), id(dimensional_calibrations[0]))
        self.assertEqual(tuple(new_xdata.dimensional_calibrations), tuple(dimensional_calibrations))
        self.assertNotEqual(id(new_xdata.metadata), id(metadata))
        self.assertEqual(new_xdata.metadata, metadata)
        self.assertNotEqual(id(new_xdata.data_descriptor), id(data_descriptor))
        self.assertEqual(new_xdata.data_descriptor, data_descriptor)

    def test_data_item_to_data_element_includes_time_zone(self):
        # created/modified are utc; timezone is specified in metadata/description/time_zone
        data_item = DataItem.DataItem(numpy.zeros((16, 16)))
        with contextlib.closing(data_item):
            data_item.created = datetime.datetime(2013, 6, 18, 14, 5, 4, 0)  # always utc
            data_item.timezone = "Europe/Athens"
            data_item.timezone_offset = "+0300"
            data_item._set_modified(datetime.datetime(2013, 6, 18, 14, 5, 4, 0))  # always utc
            data_item.metadata = {"description": {"time_zone": {"tz": "+0300", "dst": "+60"}}}
            data_element = ImportExportManager.create_data_element_from_data_item(data_item, include_data=False)
            self.assertEqual(data_element["datetime_modified"], {"dst": "+60", "local_datetime": "2013-06-18T17:05:04", 'tz': "+0300", 'timezone': "Europe/Athens"})

    def test_extended_data_to_data_element_includes_time_zone(self):
        # extended data timestamp is utc; timezone is specified in metadata/description/time_zone
        data = numpy.ones((8, 6), int)
        metadata = {"description": {"time_zone": {"tz": "+0300", "dst": "+60"}}}
        timestamp = datetime.datetime(2013, 11, 18, 14, 5, 4, 0)
        xdata = DataAndMetadata.new_data_and_metadata(data, metadata=metadata, timestamp=timestamp)
        data_element = ImportExportManager.create_data_element_from_extended_data(xdata)
        self.assertEqual(data_element["datetime_modified"], {"dst": "+60", "local_datetime": "2013-11-18T17:05:04", 'tz': "+0300"})

    def test_data_element_to_data_item_includes_time_zone(self):
        data_element = dict()
        data_element["version"] = 1
        data_element["data"] = numpy.zeros((16, 16), dtype=numpy.double)
        data_element["datetime_modified"] = {'tz': '+0300', 'dst': '+60', 'local_datetime': '2015-06-10T19:31:52.780511'}
        with contextlib.closing(ImportExportManager.create_data_item_from_data_element(data_element)) as data_item:
            self.assertEqual(data_item.timezone_offset, "+0300")
            self.assertEqual(str(data_item.created), "2015-06-10 16:31:52.780511")

    def test_data_element_to_extended_data_includes_time_zone(self):
        data_element = dict()
        data_element["version"] = 1
        data_element["data"] = numpy.zeros((16, 16), dtype=numpy.double)
        data_element["datetime_modified"] = {'tz': '+0300', 'dst': '+60', 'local_datetime': '2015-06-10T19:31:52.780511'}
        xdata = ImportExportManager.convert_data_element_to_data_and_metadata(data_element)
        self.assertEqual(xdata.timezone_offset, "+0300")
        self.assertEqual(str(xdata.timestamp), "2015-06-10 16:31:52.780511")

    def test_time_zone_in_extended_data_to_data_element_to_data_item_conversion(self):
        # test the whole path, redundant?
        data = numpy.ones((8, 6), int)
        metadata = {"description": {"time_zone": {"tz": "+0300", "dst": "+60"}}, "hardware_source": {"one": 1, "two": "b"}}
        timestamp = datetime.datetime(2013, 11, 18, 14, 5, 4, 1)
        xdata = DataAndMetadata.new_data_and_metadata(data, metadata=metadata, timestamp=timestamp)
        data_element = ImportExportManager.create_data_element_from_extended_data(xdata)
        with contextlib.closing(ImportExportManager.create_data_item_from_data_element(data_element)) as data_item:
            self.assertEqual(data_item.metadata["description"]["time_zone"]["tz"], "+0300")
            self.assertEqual(data_item.metadata["description"]["time_zone"]["dst"], "+60")
            self.assertEqual("2013-11-18 14:05:04.000001", str(data_item.created))

    def test_csv1_exporter_handles_multi_layer_display_item_with_same_calibration(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.full((50, ), 0, dtype=numpy.uint32))
            data_item.intensity_calibration = Calibration.Calibration(offset=10, scale=2)
            data_item2 = DataItem.DataItem(numpy.full((100, ), 1, dtype=numpy.uint32))
            data_item2.intensity_calibration = Calibration.Calibration(offset=5, scale=2)
            document_model.append_data_item(data_item)
            document_model.append_data_item(data_item2, False)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.display_type = "line_plot"
            display_item.append_display_data_channel_for_data_item(data_item2)
            current_working_directory = os.getcwd()
            file_path = os.path.join(current_working_directory, "__file.csv")
            handler = ImportExportManager.CSV1ImportExportHandler("csv1-io-handler", "CSV 1D", ["csv"])
            handler.write_display_item(display_item, pathlib.Path(file_path), "csv")
            self.assertTrue(os.path.exists(file_path))
            try:
                saved_data = numpy.genfromtxt(file_path, delimiter=", ")
                self.assertSequenceEqual(saved_data.shape, (max(data_item.xdata.data_shape[0], data_item2.xdata.data_shape[0]), 3))
                self.assertTrue(numpy.allclose(saved_data[:, 0], numpy.linspace(0, 99, 100)))
                self.assertTrue(numpy.allclose(saved_data[:50, 1], 10))
                self.assertTrue(numpy.allclose(saved_data[:, 2], 7))
            finally:
                os.remove(file_path)

    def test_csv1_exporter_handles_multi_layer_display_item_with_different_calibration(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.full((50, ), 0, dtype=numpy.uint32))
            data_item.dimensional_calibrations = [Calibration.Calibration(offset=0, scale=1, units="eV")]
            data_item.intensity_calibration = Calibration.Calibration(offset=10, scale=2)
            data_item2 = DataItem.DataItem(numpy.full((100, ), 1, dtype=numpy.uint32))
            data_item2.dimensional_calibrations = [Calibration.Calibration(offset=-10, scale=2, units="eV")]
            data_item2.intensity_calibration = Calibration.Calibration(offset=5, scale=2)
            document_model.append_data_item(data_item)
            document_model.append_data_item(data_item2, False)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.display_type = "line_plot"
            display_item.append_display_data_channel_for_data_item(data_item2)
            current_working_directory = os.getcwd()
            file_path = os.path.join(current_working_directory, "__file.csv")
            handler = ImportExportManager.CSV1ImportExportHandler("csv1-io-handler", "CSV 1D", ["csv"])
            handler.write_display_item(display_item, pathlib.Path(file_path), "csv")
            self.assertTrue(os.path.exists(file_path))
            try:
                saved_data = numpy.genfromtxt(file_path, delimiter=", ")
                self.assertSequenceEqual(saved_data.shape, (max(data_item.xdata.data_shape[0], data_item2.xdata.data_shape[0]), 4))
                self.assertTrue(numpy.allclose(saved_data[:50, 0], numpy.linspace(0, 49, 50)))
                self.assertTrue(numpy.allclose(saved_data[:50, 1], 10))
                self.assertTrue(numpy.allclose(saved_data[:, 2], numpy.linspace(-10, 188, 100)))
                self.assertTrue(numpy.allclose(saved_data[:, 3], 7))
            finally:
                os.remove(file_path)

    def test_data_item_to_data_element_produces_json_compatible_dict(self):
        data_item = DataItem.DataItem(numpy.zeros((16, 16)))
        with contextlib.closing(data_item):
            data_item.created = datetime.datetime(2013, 6, 18, 14, 5, 4, 0)  # always utc
            data_item.timezone = "Europe/Athens"
            data_item.timezone_offset = "+0300"
            data_item.source_file_path = "/path/to/source/file"
            data_item._set_modified(datetime.datetime(2013, 6, 18, 14, 5, 4, 0))  # always utc
            data_item.metadata = {"description": {"time_zone": {"tz": "+0300", "dst": "+60"}}}
            data_element = ImportExportManager.create_data_element_from_data_item(data_item, include_data=False)
            json.dumps(data_element)

    def test_data_item_to_data_element_and_back_keeps_large_format_flag(self):
        data_item = DataItem.DataItem(numpy.zeros((4, 4, 4)), large_format=True)
        with contextlib.closing(data_item):
            data_element = ImportExportManager.create_data_element_from_data_item(data_item, include_data=True)
            self.assertTrue(data_element.get("large_format"))
            with contextlib.closing(ImportExportManager.create_data_item_from_data_element(data_element)) as data_item:
                self.assertTrue(data_item.large_format)

    def test_importing_rgb_does_not_set_large_format(self):
        data_item = DataItem.DataItem(numpy.zeros((8, 8, 4), dtype=float))
        with contextlib.closing(data_item):
            data_item_rgb = DataItem.DataItem(numpy.zeros((8, 8, 4), dtype=numpy.uint8))
            with contextlib.closing(data_item_rgb):
                data_element = ImportExportManager.create_data_element_from_data_item(data_item, include_data=True)
                data_element_rgb = ImportExportManager.create_data_element_from_data_item(data_item_rgb, include_data=True)
                data_element.pop("large_format")
                data_element_rgb.pop("large_format")
                with contextlib.closing(ImportExportManager.create_data_item_from_data_element(data_element)) as data_item:
                    with contextlib.closing(ImportExportManager.create_data_item_from_data_element(data_element_rgb)) as data_item_rgb:
                        self.assertTrue(data_item.large_format)
                        self.assertFalse(data_item_rgb.large_format)

    def test_importing_large_numpy_file_sets_large_format_flag(self):
        current_working_directory = os.getcwd()
        file_path_npy = os.path.join(current_working_directory, "__file.npy")
        numpy.save(file_path_npy, numpy.zeros((4, 4, 4)))
        handler = ImportExportManager.NumPyImportExportHandler("numpy-io-handler", "npy", ["npy"])
        try:
            data_items = handler.read_data_items("npy", pathlib.Path(file_path_npy))
            self.assertEqual(len(data_items), 1)
            data_item = data_items[0]
            self.assertTrue(data_item.large_format)
            for data_item in data_items:
                data_item.close()
        finally:
            os.remove(file_path_npy)

    def test_data_item_with_numpy_bool_to_data_element_produces_json_compatible_dict(self):
        data_item = DataItem.DataItem(numpy.zeros((16, 16)))
        data_item.large_format = numpy.prod((2,3,4), dtype=numpy.int64) > 10  # produces a numpy.bool_
        with contextlib.closing(data_item):
            data_element = ImportExportManager.create_data_element_from_data_item(data_item, include_data=False)
            json.dumps(data_element)


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



"""Description of data model.

Reference: https://en.wikipedia.org/wiki/Data_modeling
"""

from __future__ import annotations

import copy
import gettext
import typing
import uuid

from nion.swift.model import Schema
from nion.utils import DateTime

_ = gettext.gettext

# TODO: description of physical schema
# TODO: created and modified should be implicit in all records
# TODO: should timezone, timezone offset be implicit?
# TODO: some data must be copied on read/write (dict's)
# TODO: support external data concept with loaded/unloaded property
# TODO: support loaded/unloaded entities?
# TODO: support mounted/unmounted entity concept (projects)
# TODO: support write delay / transactions
# TODO: access to auto proxy items for references
# TODO: closing, reading, inserting, removing, modifying, copying, container
# TODO: notifying, storage, resolving, moving

# TODO: are interval descriptors implicit now?
# TODO: display_layers could be a record
# TODO: display_properties could be a record
# TODO: layout could be a record
# TODO: closed_items should be a set of references

Calibration = Schema.record({
    "offset": Schema.prop(Schema.FLOAT),
    "scale": Schema.prop(Schema.FLOAT),
    "units": Schema.prop(Schema.STRING),
})

Point = Schema.fixed_tuple([Schema.prop(Schema.FLOAT), Schema.prop(Schema.FLOAT)])

Size = Schema.fixed_tuple([Schema.prop(Schema.FLOAT), Schema.prop(Schema.FLOAT)])

Rect = Schema.fixed_tuple([Point,  Size])

Vector = Schema.fixed_tuple([Point, Point])

Interval = Schema.fixed_tuple([Schema.prop(Schema.FLOAT), Schema.prop(Schema.FLOAT)])

DataItem = Schema.entity("data-item", None, 13, {
    "created": Schema.prop(Schema.TIMESTAMP),
    "data_shape": Schema.indefinite_tuple(Schema.prop(Schema.INT)),
    "data_dtype": Schema.prop(Schema.STRING),
    "is_sequence": Schema.prop(Schema.BOOLEAN),
    "collection_dimension_count": Schema.prop(Schema.INT),
    "datum_dimension_count": Schema.prop(Schema.INT),
    "intensity_calibration": Calibration,
    "dimensional_calibrations": Schema.array(Calibration),
    "data_modified": Schema.prop(Schema.TIMESTAMP),
    "timezone": Schema.prop(Schema.STRING),
    "timezone_offset": Schema.prop(Schema.STRING),
    "metadata": Schema.prop(Schema.DICT),
    "title": Schema.prop(Schema.STRING),
    "caption": Schema.prop(Schema.STRING),
    "description": Schema.prop(Schema.STRING),
    "session_id": Schema.prop(Schema.STRING),
    "session": Schema.prop(Schema.DICT),
    "category": Schema.prop(Schema.STRING, default="persistent"),
    "source": Schema.reference(),
})

DataItem.rename("source", "source_uuid")

DisplayAdjustment = Schema.entity("display_adjustment", None, None, {
})

GammaDisplayAdjustment = Schema.entity("gamma", DisplayAdjustment, None, {
    "gamma": Schema.prop(Schema.FLOAT),
})

LogDisplayAdjustment = Schema.entity("log", DisplayAdjustment, None, {
})

EqualizedDisplayAdjustment = Schema.entity("equalized", DisplayAdjustment, None, {
})

DisplayDataChannel = Schema.entity("display_data_channel", None, None, {
    "brightness": Schema.prop(Schema.FLOAT),
    "contrast": Schema.prop(Schema.FLOAT),
    "adjustments": Schema.array(Schema.component(DisplayAdjustment)),
    "complex_display_type": Schema.prop(Schema.STRING),
    "display_limits": Schema.fixed_tuple([Schema.prop(Schema.FLOAT), Schema.prop(Schema.FLOAT)]),
    "color_map_id": Schema.prop(Schema.STRING),
    "sequence_index": Schema.prop(Schema.INT),
    "collection_index": Schema.indefinite_tuple(Schema.prop(Schema.INT)),
    "slice_center": Schema.prop(Schema.INT),
    "slice_width": Schema.prop(Schema.INT),
    "data_item": Schema.reference(DataItem),
})

DisplayDataChannel.rename("data_item", "data_item_reference")

DisplayLayer = Schema.entity("display_layer", None, None, {
    "data_row": Schema.prop(Schema.INT),
    "stroke_color": Schema.prop(Schema.STRING),
    "fill_color": Schema.prop(Schema.STRING),
    "label": Schema.prop(Schema.STRING),
    "display_data_channel": Schema.reference(DisplayDataChannel),
    "stroke_width": Schema.prop(Schema.FLOAT),
})

Graphic = Schema.entity("graphic", None, None, {
    "graphic_id": Schema.prop(Schema.STRING),
    "stroke_color": Schema.prop(Schema.STRING),
    "fill_color": Schema.prop(Schema.STRING),
    "label": Schema.prop(Schema.STRING),
    "is_position_locked": Schema.prop(Schema.BOOLEAN),
    "is_shape_locked": Schema.prop(Schema.BOOLEAN),
    "is_bounds_constrained": Schema.prop(Schema.BOOLEAN),
    "role": Schema.prop(Schema.STRING),
    "source": Schema.reference(),
})

Graphic.rename("source", "source_uuid")
Graphic.rename("stroke_color", "color")

RectangleTypeGraphic = Schema.entity("rect-type-graphic", Graphic, None, {
    "bounds": Rect,
    "rotation": Schema.prop(Schema.FLOAT),
})

RectangleGraphic = Schema.entity("rect-graphic", RectangleTypeGraphic, None, {})

EllipseGraphic = Schema.entity("ellipse-graphic", RectangleTypeGraphic, None, {})

LineTypeGraphic = Schema.entity("line-type-graphic", Graphic, None, {
    "start": Point,
    "end": Point,
    "start_arrow_enabled": Schema.prop(Schema.BOOLEAN),
    "end_arrow_enabled": Schema.prop(Schema.BOOLEAN),
})

LineGraphic = Schema.entity("line-graphic", LineTypeGraphic, None, {})

LineProfileGraphic = Schema.entity("line-profile-graphic", LineTypeGraphic, None, {
    "width": Schema.prop(Schema.FLOAT, default=1.0),
    "interval_descriptors": Schema.prop(Schema.LIST),
})

PointGraphic = Schema.entity("point-graphic", Graphic, None, {
    "position": Point,
})

IntervalGraphic = Schema.entity("interval-graphic", Graphic, None, {
    "start": Schema.prop(Schema.FLOAT, default=0.0),
    "end": Schema.prop(Schema.FLOAT, default=1.0),
})

ChannelGraphic = Schema.entity("channel-graphic", Graphic, None, {
    "position": Schema.prop(Schema.FLOAT, default=0.5),
})

SpotGraphic = Schema.entity("spot-graphic", Graphic, None, {
    "bounds": Rect,
    "rotation": Schema.prop(Schema.FLOAT),
})

WedgeGraphic = Schema.entity("wedge-graphic", Graphic, None, {
    "angle_interval": Interval,
})

RingGraphic = Schema.entity("ring-graphic", Graphic, None, {
    "radius_1": Schema.prop(Schema.FLOAT, default=0.2),
    "radius_2": Schema.prop(Schema.FLOAT, default=0.4),
    "mode": Schema.prop(Schema.STRING, default="band-pass"),
})

LatticeGraphic = Schema.entity("lattice-graphic", Graphic, None, {
    "u_pos": Point,
    "v_pos": Point,
    "u_count": Schema.prop(Schema.INT, default=1),
    "v_count": Schema.prop(Schema.INT, default=1),
    "radius": Schema.prop(Schema.FLOAT, default=0.1),
})

DisplayItem = Schema.entity("display_item", None, None, {
    "created": Schema.prop(Schema.TIMESTAMP),
    "display_type": Schema.prop(Schema.STRING),
    "title": Schema.prop(Schema.STRING),
    "caption": Schema.prop(Schema.STRING),
    "description": Schema.prop(Schema.STRING),
    "session_id": Schema.prop(Schema.STRING),
    "calibration_style_id": Schema.prop(Schema.STRING, default="calibrated"),
    "display_properties": Schema.prop(Schema.DICT),
    "display_layers": Schema.array(Schema.component(DisplayLayer)),
    "graphics": Schema.array(Schema.component(Graphic)),
    "display_data_channels": Schema.array(Schema.component(DisplayDataChannel)),
})

Specifier = Schema.entity("specifier", None, None, {
    "version": Schema.prop(Schema.INT),
    "reference": Schema.reference(),
})

Specifier.rename("reference", "reference_uuid")

EmptySpecifier = Schema.entity("empty_specifier", Specifier, None, {})
DataSourceSpecifier = Schema.entity("data_source", Specifier, None, {})
DataItemSpecifier = Schema.entity("data_item", Specifier, None, {})
GraphicSpecifier = Schema.entity("graphic-specifier", Specifier, None, {})
StructureSpecifier = Schema.entity("structure", Specifier, None, {})
DataSpecifier = Schema.entity("xdata", Specifier, None, {})
DisplayDataSpecifier = Schema.entity("display_xdata", Specifier, None, {})
CroppedDataSpecifier = Schema.entity("cropped_xdata", Specifier, None, {})
CroppedDisplayDataSpecifier = Schema.entity("cropped_display_xdata", Specifier, None, {})
FilterDataSpecifier = Schema.entity("filter_xdata", Specifier, None, {})
FilteredDataSpecifier = Schema.entity("filtered_xdata", Specifier, None, {})

ComputationVariable = Schema.entity("variable", None, None, {
    "name": Schema.prop(Schema.STRING),
    "label": Schema.prop(Schema.STRING),
    "value_type": Schema.prop(Schema.STRING),
    "value": Schema.prop(Schema.ANY),
    "value_default": Schema.prop(Schema.ANY),
    "value_min": Schema.prop(Schema.ANY),
    "value_max": Schema.prop(Schema.ANY),
    "item": Schema.component(Specifier, required=False),
    "item2": Schema.component(Specifier, required=False),
    "items": Schema.array(Schema.component(Specifier), Schema.OPTIONAL),
    "property_name": Schema.prop(Schema.STRING),
    "control_type": Schema.prop(Schema.STRING),
})

ComputationVariable.rename("item", "specifier")
ComputationVariable.rename("item2", "secondary_specifier")
ComputationVariable.rename("items", "object_specifiers")

ComputationResult = Schema.entity("output", None, None, {
    "name": Schema.prop(Schema.STRING),
    "label": Schema.prop(Schema.STRING),
    "item": Schema.component(Specifier),
    "items": Schema.array(Schema.component(Specifier), Schema.OPTIONAL),
})

ComputationResult.rename("item", "specifier")
ComputationResult.rename("items", "specifiers")

Computation = Schema.entity("computation", None, None, {
    "source": Schema.reference(),
    "original_expression": Schema.prop(Schema.STRING),
    "error_text": Schema.prop(Schema.STRING),
    "label": Schema.prop(Schema.STRING),
    "processing_id": Schema.prop(Schema.STRING),
    "variables": Schema.array(Schema.component(ComputationVariable)),
    "results": Schema.array(Schema.component(ComputationResult)),
})

Computation.rename("source", "source_uuid")

DataStructure = Schema.entity("data_structure", None, None, {
    "source": Schema.reference(),
    "structure_type": Schema.prop(Schema.STRING),
    "properties": Schema.prop(Schema.DICT),
})

DataStructure.rename("source", "source_uuid")

Connection = Schema.entity("connection", None, None, {
    "parent": Schema.reference(),
})

Connection.rename("parent", "parent_uuid")

PropertyConnection = Schema.entity("property-connection", Connection, None, {
    "source": Schema.reference(),
    "source_property": Schema.prop(Schema.STRING),
    "target": Schema.reference(),
    "target_property": Schema.prop(Schema.STRING),
})

PropertyConnection.rename("source", "source_uuid")
PropertyConnection.rename("target", "target_uuid")

IntervalListConnection = Schema.entity("interval-list-connection", Connection, None, {
    "source": Schema.reference(),
    "target": Schema.reference(),
})

IntervalListConnection.rename("source", "source_uuid")
IntervalListConnection.rename("target", "target_uuid")

DataGroup = Schema.entity("data_group", None, None, {
    "title": Schema.prop(Schema.STRING, default=_("Untitled")),
    "display_items": Schema.array(Schema.component(DisplayItem)),
    "data_groups": Schema.array(Schema.component("data_group")),
})

DataGroup.rename("display_items", "display_item_references")

Workspace = Schema.entity("workspace", None, None, {
    "name": Schema.prop(Schema.STRING),
    "created": Schema.prop(Schema.TIMESTAMP),
    "layout": Schema.prop(Schema.DICT, Schema.OPTIONAL),
    "workspace_id": Schema.prop(Schema.STRING, Schema.OPTIONAL),
})

Project = Schema.entity("project", None, 3, {
    "title": Schema.prop(Schema.STRING),
    "data_items": Schema.array(Schema.component(DataItem), Schema.OPTIONAL),
    "display_items": Schema.array(Schema.component(DisplayItem)),
    "computations": Schema.array(Schema.component(Computation)),
    "data_structures": Schema.array(Schema.component(DataStructure)),
    "connections": Schema.array(Schema.component(Connection)),
    "data_groups": Schema.array(Schema.component(DataGroup)),
    "workspaces": Schema.array(Schema.component(Workspace)),
    "workspace": Schema.reference(Workspace),
    "data_item_references": Schema.map(Schema.STRING, Schema.reference(DataItem)),
    "mapped_items": Schema.array(Schema.reference(DataItem), Schema.OPTIONAL),
    "project_data_folders": Schema.array(Schema.prop(Schema.PATH)),
})

Project.rename("workspace", "workspace_uuid")

PersistentDictType = typing.Dict[str, typing.Any]


def transform_forward(d: PersistentDictType) -> PersistentDictType:
    # ensure the display_layer has a uuid and modified and looks like a regular entity.
    for display_item in d.get("display_items", list()):
        display_data_channels = display_item.get("display_data_channels", list())
        for display_layer in display_item.get("display_layers", list()):
            display_layer["type"] = "display_layer"
            display_layer["uuid"] = str(uuid.uuid4())
            display_layer["modified"] = copy.copy(display_item.get("modified", DateTime.utcnow().isoformat()))
            data_index = display_layer.pop("data_index", None)
            if data_index is not None and 0 <= data_index < len(display_data_channels):
                display_layer["display_data_channel"] = display_data_channels[data_index]["uuid"]

    # ensure the specifier to graphic has a unique entity_id
    # note this uses the non-renamed fields (specifier=item; secondary_specifier=item2; object_specifiers=items; specifiers=items)
    # also change uuid fields to reference_uuid
    for computation_d in d.get("computations", list()):
        for variable_d in computation_d.get("variables", list()):
            specifier_d = variable_d.get("specifier", dict())
            secondary_specifier_d = variable_d.get("secondary_specifier", dict())
            # clean up secondary_specifier
            if not secondary_specifier_d:
                variable_d.pop("secondary_specifier", None)
            if specifier_d:
                specifier_d["reference_uuid"] = specifier_d.pop("uuid")
            if secondary_specifier_d:
                secondary_specifier_d["reference_uuid"] = secondary_specifier_d.pop("uuid")
            # now update the graphic types
            if specifier_d.get("type") == "graphic":
                variable_d["specifier"]["type"] = "graphic-specifier"
            if secondary_specifier_d.get("type") == "graphic":
                variable_d["secondary_specifier"]["type"] = "graphic-specifier"
            for v in variable_d.get("object_specifiers", list()):
                if v is not None:
                    v["reference_uuid"] = v.pop("uuid")
                    if v.get("type") == "graphic":
                        v["type"] = "graphic-specifier"
        for result_d in computation_d.get("results", list()):
            result_specifier_d = result_d.get("specifier", None)
            if result_specifier_d is not None:
                result_specifier_d["reference_uuid"] = result_specifier_d.pop("uuid")
                if result_specifier_d.get("type") == "graphic":
                    result_specifier_d["type"] = "graphic-specifier"
            for r in result_d.get("specifiers", list()):
                if r is not None:
                    r["reference_uuid"] = r.pop("uuid")
                    if r.get("type") == "graphic":
                        r["type"] = "graphic-specifier"

    return d


def transform_backward(d: PersistentDictType) -> PersistentDictType:
    # ensure the display_layer has a uuid and modified and looks like a regular entity (reverse)
    for display_item in d.get("display_items", list()):
        display_data_channels = display_item.get("display_data_channels", list())
        display_data_channel_map = {display_data_channel["uuid"]: index for index, display_data_channel in enumerate(display_data_channels)}
        for display_layer in display_item.get("display_layers", list()):
            display_layer.pop("type", None)
            display_layer.pop("uuid", None)
            display_layer.pop("modified", None)
            display_data_channel_uuid = display_layer.pop("display_data_channel", None)
            data_index = display_data_channel_map.get(display_data_channel_uuid, None)
            if data_index is not None:
                display_layer["data_index"] = data_index

    # ensure the specifier to graphic uses PROJECT_VERSION 3 compatible keys
    # note this uses the non-renamed fields (specifier=item; secondary_specifier=item2; object_specifiers=items; specifiers=items)

    # NOTE: The backing dict is not carefully handled - so sometimes it will contain the transformed dict and sometimes
    # it won't. Obviously this is a problem; but the code currently works around it by adding extra checks for whether
    # reference_uuid is in the item dict's. This will be cleaned up in a future refactoring.

    for computation_d in d.get("computations", list()):
        for variable_d in computation_d.get("variables", list()):
            specifier_d = variable_d.get("specifier", dict())
            secondary_specifier_d = variable_d.get("secondary_specifier", dict())
            if specifier_d and "reference_uuid" in specifier_d:
                specifier_d["uuid"] = specifier_d.pop("reference_uuid")
            if secondary_specifier_d and "reference_uuid" in secondary_specifier_d:
                secondary_specifier_d["uuid"] = secondary_specifier_d.pop("reference_uuid")
            if specifier_d.get("type") == "graphic-specifier":
                variable_d["specifier"]["type"] = "graphic"
            if secondary_specifier_d.get("type") == "graphic-specifier":
                variable_d["secondary_specifier"]["type"] = "graphic"
            for v in variable_d.get("object_specifiers", list()):
                if v is not None:
                    if "reference_uuid" in v:
                        v["uuid"] = v.pop("reference_uuid")
                    if v.get("type") == "graphic-specifier":
                        v["type"] = "graphic"
        for result_d in computation_d.get("results", list()):
            result_specifier_d = result_d.get("specifier", dict())
            if result_specifier_d and "reference_uuid" in result_specifier_d:
                result_specifier_d["uuid"] = result_specifier_d.pop("reference_uuid")
            if result_specifier_d.get("type") == "graphic-specifier":
                result_specifier_d["type"] = "graphic"
            for r in result_d.get("specifiers", list()):
                if r is not None:
                    if "reference_uuid" in r:
                        r["uuid"] = r.pop("reference_uuid")
                    if r.get("type") == "graphic-specifier":
                        r["type"] = "graphic"

    return d


Project.transform(transform_forward, transform_backward)

ProjectReference = Schema.entity("project_reference", None, None, {
    "project": Schema.reference(Project),
    "is_active": Schema.prop(Schema.BOOLEAN),
    "last_used": Schema.prop(Schema.TIMESTAMP),
})

ProjectReference.rename("project", "project_uuid")

IndexProjectReference = Schema.entity("project_index", ProjectReference, None, {
    "project_path": Schema.prop(Schema.PATH),
})

FolderProjectReference = Schema.entity("project_folder", ProjectReference, None, {
    "project_folder_path": Schema.prop(Schema.PATH),
})

MemoryProjectReference = Schema.entity("project_memory", ProjectReference, None, {
})

ScriptItem = Schema.entity("script_item", None, None, {
})

FileScriptItem = Schema.entity("file_script_item", ScriptItem, None, {
    "path": Schema.prop(Schema.PATH),
})

FolderScriptItem = Schema.entity("folder_script_item", ScriptItem, None, {
    "folder_path": Schema.prop(Schema.PATH),
    "is_closed": Schema.prop(Schema.BOOLEAN),
})

Profile = Schema.entity("profile", None, 2, {
    "project_references": Schema.array(Schema.component(ProjectReference)),
    "last_project_reference": Schema.reference(ProjectReference),
    "work_project": Schema.reference(ProjectReference),
    "closed_items": Schema.prop(Schema.SET),
    "script_items": Schema.array(Schema.component(ScriptItem)),
    "script_items_updated": Schema.prop(Schema.BOOLEAN)
})

Profile.rename("target_project", "target_project_reference_uuid")
Profile.rename("work_project", "work_project_reference_uuid")



import argparse
import os
import subprocess
import sys
import typing

parser = argparse.ArgumentParser(
    prog="python -m nionswift", description="Launch Nion Swift using native launcher or pyside6 libraries."
)

parser.add_argument(
    "--ui",
    dest="ui",
    action="store",
    choices=["tool", "qt"],
    help="choose UI frontend",
    default="tool",
)

parser.add_argument(
    "--fallback",
    dest="fallback",
    action=argparse.BooleanOptionalAction,
    help="whether to fall back to other UI frontend if preferred choice is unavailable",
    default=True,
)

parsed_args = parser.parse_args()

app_id = "nionui_app.nionswift"
args = list[typing.Any]()

order = list[str]()

if parsed_args.ui:
    order.append(parsed_args.ui)

# if using fallback, add tool and qt to order if not already present.
if parsed_args.fallback:
    if "tool" not in order:
        order.append("tool")
    if "qt" not in order:
        order.append("qt")

# go through the ui preferences in order
for ui in order:
    if ui == "qt":
        # launch the app using the pyside6 qt frontend
        from nionui_app.nionswift import main
        app = main.main(list(), {"qt": None})
        app.run()
        break
    else:
        # launch using the tool frontend
        if sys.platform == "darwin":
            exe_path = os.path.join(sys.exec_prefix, "bin", "Nion Swift.app", "Contents", "MacOS", "Nion Swift")
        elif sys.platform == "linux":
            exe_path = os.path.join(sys.exec_prefix, "bin", "NionSwiftLauncher", "NionSwiftLauncher")
        elif sys.platform == "win32":
            exe_path = os.path.join(sys.exec_prefix, "Scripts", "NionSwiftLauncher", "NionSwift.exe")
        else:
            exe_path = None
        if exe_path:
            python_prefix = sys.prefix
            proc = subprocess.Popen([exe_path, python_prefix, app_id] + args, universal_newlines=True)
            proc.communicate()
            break



# standard libraries
import abc
import functools
import typing

# third party libraries
# None

# local libraries
from nion.utils import Converter
from nion.utils import Event
from nion.utils import Observable


class PropertyConnectorItem:

    def __init__(self, item: Observable.Observable, property_name: str, converter: typing.Optional[Converter.ConverterLike[typing.Any, typing.Any]] = None) -> None:
        self.item = item
        self.property_name = property_name
        self.converter = converter

    def get_common_value(self) -> typing.Any:
        value = getattr(self.item, self.property_name)
        return self.converter.convert(value) if self.converter else value

    def set_common_value(self, formatted_value: typing.Any) -> None:
        value = self.converter.convert_back(formatted_value) if self.converter else formatted_value
        if getattr(self.item, self.property_name) != value:
            setattr(self.item, self.property_name, value)


class PropertyConnector:
    # the converter converts from the internal property to common property.
    # when a property change is received, it converts back to the common property.
    # when a common property is sent out to other items it is converted from the
    # common property.

    def __init__(self, property_connector_items: typing.Sequence[PropertyConnectorItem]) -> None:
        self.__property_changed_listeners: typing.List[Event.EventListener] = list()

        self.__suppress = False

        def property_changed(property_connector_item: PropertyConnectorItem, key: str) -> None:
            if not self.__suppress:
                self.__suppress = True
                try:
                    if key == property_connector_item.property_name:
                        common_value = property_connector_item.get_common_value()
                        for i_property_connector_item in property_connector_items:
                            if i_property_connector_item != property_connector_item:
                                i_property_connector_item.set_common_value(common_value)
                finally:
                    self.__suppress = False

        for property_connector_item in property_connector_items:
            self.__property_changed_listeners.append(property_connector_item.item.property_changed_event.listen(functools.partial(property_changed, property_connector_item)))

        property_changed(property_connector_items[0], property_connector_items[0].property_name)

    def close(self) -> None:
        for listener in self.__property_changed_listeners:
            listener.close()
        self.__property_changed_listeners = typing.cast(typing.List[Event.EventListener], None)



from __future__ import annotations

# standard libraries
import functools
import logging
import operator
import time
import typing
import uuid
import weakref

# local libraries
from nion.swift.model import Cache
from nion.swift.model import Changes
from nion.swift.model import Connection
from nion.swift.model import DataGroup
from nion.swift.model import Symbolic
from nion.swift.model import DataItem
from nion.swift.model import DataStructure
from nion.swift.model import DisplayItem
from nion.swift.model import FileStorageSystem
from nion.swift.model import Persistence
from nion.swift.model import WorkspaceLayout
from nion.utils import Converter
from nion.utils import ListModel

PersistentDictType = typing.Dict[str, typing.Any]

ProjectItemType = typing.Union[DataItem.DataItem, DisplayItem.DisplayItem, DataStructure.DataStructure, Connection.Connection, Symbolic.Computation]


class Project(Persistence.PersistentObject):
    """A project manages raw data items, display items, computations, data structures, and connections.

    Projects are stored in project indexes, which are files that describe how to find data and and tracks the other
    project relationships (display items, computations, data structures, connections).

    Projects manage reading, writing, and data migration.
    """

    PROJECT_VERSION = 3

    _processing_descriptions: PersistentDictType = dict()

    def __init__(self, storage_system: Persistence.PersistentStorageInterface, cache_factory: typing.Optional[Cache.CacheFactory] = None) -> None:
        super().__init__()

        self.define_type("project")
        self.define_property("title", str(), hidden=True)
        self.define_relationship("data_items", data_item_factory, insert=self.__data_item_inserted, remove=self.__data_item_removed, hidden=True)
        self.define_relationship("display_items", display_item_factory, insert=self.__display_item_inserted, remove=self.__display_item_removed, hidden=True)
        self.define_relationship("computations", computation_factory, insert=self.__computation_inserted, remove=self.__computation_removed, hidden=True)
        self.define_relationship("data_structures", data_structure_factory, insert=self.__data_structure_inserted, remove=self.__data_structure_removed, hidden=True)
        self.define_relationship("connections", Connection.connection_factory, insert=self.__connection_inserted, remove=self.__connection_removed, hidden=True)
        self.define_relationship("data_groups", DataGroup.data_group_factory, insert=self.__data_group_inserted, remove=self.__data_group_removed, hidden=True)
        self.define_relationship("workspaces", WorkspaceLayout.factory, hidden=True)
        self.define_property("workspace_uuid", converter=Converter.UuidToStringConverter(), hidden=True)
        self.define_property("data_item_references", dict(), changed=self.__property_changed, hidden=True)  # map string key to data item, used for data acquisition channels
        self.define_property("mapped_items", list(), changed=self.__property_changed, hidden=True)  # list of item references, used for shortcut variables in scripts

        self.handle_start_read: typing.Optional[typing.Callable[[], None]] = None
        self.handle_insert_model_item: typing.Optional[typing.Callable[[Persistence.PersistentContainerType, str, int, Persistence.PersistentObject], None]] = None
        self.handle_remove_model_item: typing.Optional[typing.Callable[[Persistence.PersistentContainerType, str, Persistence.PersistentObject, bool], Changes.UndeleteLog]] = None
        self.handle_finish_read: typing.Optional[typing.Callable[[], None]] = None

        self.__has_been_read = False
        self.__project_load_start_time = 0.0

        self._raw_properties: typing.Optional[PersistentDictType] = None
        self.__reader_errors: typing.Sequence[Persistence.ReaderError] = list()

        self.__storage_system = storage_system

        self.set_storage_system(self.__storage_system)

        self.__cache_factory = cache_factory
        self.__cache = cache_factory.create_cache() if cache_factory else None

    def close(self) -> None:
        self.handle_start_read = None
        self.handle_insert_model_item = None
        self.handle_remove_model_item = None
        self.handle_finish_read = None
        if self.__cache_factory:
            self.__cache_factory.release_cache(typing.cast(Cache.CacheLike, self.__cache))
            self.__cache_factory = None
            self.__cache = None
        self.__storage_system.close()
        self.__storage_system = typing.cast(typing.Any, None)
        super().close()

    @property
    def title(self) -> str:
        return typing.cast(str, self._get_persistent_property_value("title"))

    @title.setter
    def title(self, value: str) -> None:
        self._set_persistent_property_value("title", value)

    @property
    def workspace_uuid(self) -> typing.Optional[uuid.UUID]:
        uuid_str = typing.cast(typing.Optional[str], self._get_persistent_property_value("workspace_uuid"))
        return uuid.UUID(uuid_str) if uuid_str else None

    @workspace_uuid.setter
    def workspace_uuid(self, value: typing.Optional[uuid.UUID]) -> None:
        self._set_persistent_property_value("workspace_uuid", str(value) if value else None)

    @property
    def mapped_items(self) -> typing.List[Persistence._SpecifierType]:
        return list(self._get_persistent_property_value("mapped_items"))

    @mapped_items.setter
    def mapped_items(self, value: typing.List[Persistence._SpecifierType]) -> None:
        self._set_persistent_property_value("mapped_items", value)

    @property
    def data_items(self) -> typing.Sequence[DataItem.DataItem]:
        return typing.cast(typing.Sequence[DataItem.DataItem], self._get_relationship_values("data_items"))

    @property
    def display_items(self) -> typing.Sequence[DisplayItem.DisplayItem]:
        return typing.cast(typing.Sequence[DisplayItem.DisplayItem], self._get_relationship_values("display_items"))

    @property
    def computations(self) -> typing.Sequence[Symbolic.Computation]:
        return typing.cast(typing.Sequence[Symbolic.Computation], self._get_relationship_values("computations"))

    @property
    def data_structures(self) -> typing.Sequence[DataStructure.DataStructure]:
        return typing.cast(typing.Sequence[DataStructure.DataStructure], self._get_relationship_values("data_structures"))

    @property
    def connections(self) -> typing.Sequence[Connection.Connection]:
        return typing.cast(typing.Sequence[Connection.Connection], self._get_relationship_values("connections"))

    @property
    def data_groups(self) -> typing.Sequence[DataGroup.DataGroup]:
        return typing.cast(typing.Sequence[DataGroup.DataGroup], self._get_relationship_values("data_groups"))

    @property
    def workspaces(self) -> typing.Sequence[WorkspaceLayout.WorkspaceLayout]:
        return typing.cast(typing.Sequence[WorkspaceLayout.WorkspaceLayout], self._get_relationship_values("workspaces"))

    @property
    def sorted_workspaces(self) -> typing.Sequence[WorkspaceLayout.WorkspaceLayout]:
        return sorted(self.workspaces, key=operator.attrgetter("timestamp_for_sorting"), reverse=True)

    def open(self) -> None:
        self.__storage_system.reset()  # this makes storage reusable during tests

    def create_proxy(self) -> Persistence.PersistentObjectProxy[Project]:
        container = self.container
        assert container
        return container.create_item_proxy(item=self)

    @property
    def item_specifier(self) -> Persistence.PersistentObjectSpecifier:
        return Persistence.PersistentObjectSpecifier(self.uuid)

    def insert_model_item(self, container: Persistence.PersistentContainerType, name: str, before_index: int, item: Persistence.PersistentObject) -> None:
        # special handling to pass on to the document model
        assert callable(self.handle_insert_model_item)
        self.handle_insert_model_item(container, name, before_index, item)

    def remove_model_item(self, container: Persistence.PersistentContainerType, name: str, item: Persistence.PersistentObject, *, safe: bool = False) -> Changes.UndeleteLog:
        # special handling to pass on to the document model
        assert callable(self.handle_remove_model_item)
        return self.handle_remove_model_item(container, name, item, safe)

    @property
    def storage_location_str(self) -> str:
        return self.__storage_system.get_identifier()

    @property
    def storage_cache(self) -> Cache.CacheLike:
        assert self.__cache
        return self.__cache

    @property
    def project_uuid(self) -> typing.Optional[uuid.UUID]:
        properties = self.__storage_system.get_storage_properties()
        try:
            return uuid.UUID(properties.get("uuid", str(uuid.uuid4()))) if properties else None
        except Exception as e:
            return None

    @property
    def project_state(self) -> str:
        properties = self.__storage_system.get_storage_properties()
        if properties is not None and not properties:
            return "missing"
        project_uuid = self.project_uuid
        project_version = self.project_version
        if project_uuid is not None and project_version is not None:
            if project_version  == FileStorageSystem.PROJECT_VERSION:
                return "loaded" if self.__has_been_read else "unloaded"
            else:
                return "needs_upgrade"
        return "invalid"

    @property
    def project_version(self) -> typing.Optional[int]:
        properties = self.__storage_system.get_storage_properties()
        try:
            return properties.get("version", None) if properties else None
        except Exception:
            return None

    @property
    def project_filter(self) -> ListModel.Filter:

        # Python 3.9+ weak ref
        def is_display_item_active(project_weak_ref: typing.Any, display_item: DisplayItem.DisplayItem) -> bool:
            return bool(display_item.project == project_weak_ref())

        # use a weak reference to avoid circular references loops that prevent garbage collection
        return ListModel.PredicateFilter(functools.partial(is_display_item_active, weakref.ref(self)))

    @property
    def project_storage_system(self) -> Persistence.PersistentStorageInterface:
        return self.__storage_system

    def __data_item_inserted(self, name: str, before_index: int, data_item: DataItem.DataItem) -> None:
        self.notify_insert_item("data_items", data_item, before_index)

    def __data_item_removed(self, name: str, index: int, data_item: DataItem.DataItem) -> None:
        self.notify_remove_item("data_items", data_item, index)

    def __display_item_inserted(self, name: str, before_index: int, display_item: DisplayItem.DisplayItem) -> None:
        self.notify_insert_item("display_items", display_item, before_index)

    def __display_item_removed(self, name: str, index: int, display_item: DisplayItem.DisplayItem) -> None:
        self.notify_remove_item("display_items", display_item, index)

    def __data_structure_inserted(self, name: str, before_index: int, data_structure: DataStructure.DataStructure) -> None:
        self.notify_insert_item("data_structures", data_structure, before_index)

    def __data_structure_removed(self, name: str, index: int, data_structure: DataStructure.DataStructure) -> None:
        self.notify_remove_item("data_structures", data_structure, index)

    def __computation_inserted(self, name: str, before_index: int, computation: Symbolic.Computation) -> None:
        self.notify_insert_item("computations", computation, before_index)

    def __computation_removed(self, name: str, index: int, computation: Symbolic.Computation) -> None:
        self.notify_remove_item("computations", computation, index)

    def __connection_inserted(self, name: str, before_index: int, connection: Connection.Connection) -> None:
        self.notify_insert_item("connections", connection, before_index)

    def __connection_removed(self, name: str, index: int, connection: Connection.Connection) -> None:
        self.notify_remove_item("connections", connection, index)

    def __data_group_inserted(self, name: str, before_index: int, data_group: DataGroup.DataGroup) -> None:
        self.notify_insert_item("data_groups", data_group, before_index)

    def __data_group_removed(self, name: str, index: int, data_group: DataGroup.DataGroup) -> None:
        self.notify_remove_item("data_groups", data_group, index)

    def prepare_read_project(self) -> None:
        self.__project_load_start_time = time.time()
        logging.getLogger("loader").info(f"Loading project {self.__storage_system.get_identifier()}")
        self._raw_properties, self.__reader_errors = self.__storage_system.read_project_properties()  # combines library and data item properties
        self.uuid = uuid.UUID(self._raw_properties.get("uuid", str(uuid.uuid4())))

        for reader_error in self.__reader_errors:
            logging.getLogger("loader").error(f"Error reading {reader_error.identifier} ({reader_error.exception})")
            if False and reader_error.tb is not None:
                import sys
                import traceback
                print(f"Event Listener Traceback (most recent call last)", file=sys.stderr)
                frame_summaries = typing.cast(typing.List[typing.Any], reader_error.tb)
                for line in traceback.StackSummary.from_list(frame_summaries).format():
                    print(line, file=sys.stderr, end="")

    def read_project(self) -> None:
        if callable(self.handle_start_read):
            self.handle_start_read()
        properties = self._raw_properties
        if properties:
            project_version = properties.get("version", None)
            if project_version is not None and project_version == FileStorageSystem.PROJECT_VERSION:
                for item_d in properties.get("data_items", list()):
                    data_item = DataItem.DataItem()
                    data_item.begin_reading()
                    data_item.read_from_dict(item_d)
                    data_item.finish_reading()
                    if not self.get_item_by_uuid("data_items", data_item.uuid):
                        self.load_item("data_items", len(self.data_items), data_item)
                    else:
                        data_item.close()
                for item_d in properties.get("display_items", list()):
                    display_item = DisplayItem.DisplayItem()
                    display_item.begin_reading()
                    display_item.read_from_dict(item_d)
                    display_item.finish_reading()
                    if not self.get_item_by_uuid("display_items", display_item.uuid):
                        self.load_item("display_items", len(self.display_items), display_item)
                    else:
                        display_item.close()
                for item_d in properties.get("data_structures", list()):
                    data_structure = DataStructure.DataStructure()
                    data_structure.begin_reading()
                    data_structure.read_from_dict(item_d)
                    data_structure.finish_reading()
                    if not self.get_item_by_uuid("data_structures", data_structure.uuid):
                        self.load_item("data_structures", len(self.data_structures), data_structure)
                    else:
                        data_structure.close()
                for item_d in properties.get("computations", list()):
                    computation = Symbolic.Computation()
                    computation.begin_reading()
                    computation.read_from_dict(item_d)
                    computation.finish_reading()
                    if not self.get_item_by_uuid("computations", computation.uuid):
                        self.load_item("computations", len(self.computations), computation)
                        # TODO: handle update script and bind after reload in document model
                        computation.update_script(Project._processing_descriptions)
                        computation.reset()
                    else:
                        computation.close()
                for item_d in properties.get("connections", list()):
                    connection = Connection.connection_factory(item_d.get)
                    if connection:
                        connection.begin_reading()
                        connection.read_from_dict(item_d)
                        connection.finish_reading()
                        if not self.get_item_by_uuid("connections", connection.uuid):
                            self.load_item("connections", len(self.connections), connection)
                        else:
                            connection.close()
                for item_d in properties.get("data_groups", list()):
                    data_group = DataGroup.data_group_factory(item_d.get)
                    if data_group:
                        data_group.begin_reading()
                        data_group.read_from_dict(item_d)
                        data_group.finish_reading()
                        if not self.get_item_by_uuid("data_groups", data_group.uuid):
                            self.load_item("data_groups", len(self.data_groups), data_group)
                        else:
                            data_group.close()
                for item_d in properties.get("workspaces", list()):
                    workspace = WorkspaceLayout.factory(item_d.get)
                    workspace.begin_reading()
                    workspace.read_from_dict(item_d)
                    workspace.finish_reading()
                    if not self.get_item_by_uuid("workspaces", workspace.uuid):
                        self.load_item("workspaces", len(self.workspaces), workspace)
                    else:
                        workspace.close()
                workspace_uuid_str = properties.get("workspace_uuid", None)
                workspace_uuid = uuid.UUID(workspace_uuid_str) if workspace_uuid_str else None
                existing_workspace_uuid = self._get_persistent_property_value("workspace_uuid", None)
                if workspace_uuid and existing_workspace_uuid != workspace_uuid:
                    self._get_persistent_property("workspace_uuid").set_value(str(workspace_uuid))
                self._get_persistent_property("data_item_references").set_value(properties.get("data_item_references", dict()))
                self._get_persistent_property("mapped_items").set_value(properties.get("mapped_items", list()))
                self.__has_been_read = True
        if callable(self.handle_finish_read):
            self.handle_finish_read()
        elapsed_s = int(time.time() - self.__project_load_start_time)
        logging.getLogger("loader").info(f"Loaded project {elapsed_s}s (data items: {len(self.data_items)}, display items: {len(self.display_items)}, data structures: {len(self.data_structures)}, computations: {len(self.computations)}, connections: {len(self.connections)}, data groups: {len(self.data_groups)}, workspaces: {len(self.workspaces)})")

    def __property_changed(self, name: str, value: typing.Any) -> None:
        self.notify_property_changed(name)

    def append_data_item(self, data_item: DataItem.DataItem) -> None:
        assert not self.get_item_by_uuid("data_items", data_item.uuid)
        self.append_item("data_items", data_item)
        data_item.write_data_if_not_delayed()  # initially write to disk

    def remove_data_item(self, data_item: DataItem.DataItem) -> None:
        self.remove_item("data_items", data_item)

    def restore_data_item(self, data_item_uuid: uuid.UUID) -> typing.Optional[DataItem.DataItem]:
        item_d = self.__storage_system.restore_item(data_item_uuid)
        if item_d is not None:
            data_item_uuid = uuid.UUID(item_d.get("uuid"))
            large_format = item_d.get("__large_format", False)
            data_item = DataItem.DataItem(item_uuid=data_item_uuid, large_format=large_format)
            data_item.begin_reading()
            data_item.read_from_dict(item_d)
            data_item.finish_reading()
            assert not self.get_item_by_uuid("data_items", data_item.uuid)
            self.append_item("data_items", data_item)
            assert data_item.container == self
            return data_item
        return None

    def append_display_item(self, display_item: DisplayItem.DisplayItem) -> None:
        assert not self.get_item_by_uuid("display_items", display_item.uuid)
        self.append_item("display_items", display_item)

    def remove_display_item(self, display_item: DisplayItem.DisplayItem) -> None:
        self.remove_item("display_items", display_item)

    def append_data_structure(self, data_structure: DataStructure.DataStructure) -> None:
        assert not self.get_item_by_uuid("data_structures", data_structure.uuid)
        self.append_item("data_structures", data_structure)

    def remove_data_structure(self, data_structure: DataStructure.DataStructure) -> None:
        self.remove_item("data_structures", data_structure)

    def append_computation(self, computation: Symbolic.Computation) -> None:
        assert not self.get_item_by_uuid("computations", computation.uuid)
        self.append_item("computations", computation)

    def remove_computation(self, computation: Symbolic.Computation) -> None:
        self.remove_item("computations", computation)

    def append_connection(self, connection: Connection.Connection) -> None:
        assert not self.get_item_by_uuid("connections", connection.uuid)
        self.append_item("connections", connection)

    def remove_connection(self, connection: Connection.Connection) -> None:
        self.remove_item("connections", connection)

    @property
    def data_item_references(self) -> typing.Dict[str, str]:
        return dict(self._get_persistent_property_value("data_item_references").items())

    def set_data_item_reference(self, key: str, data_item: DataItem.DataItem) -> None:
        data_item_references = self.data_item_references
        data_item_references[key] = str(data_item.item_specifier.write())
        self._set_persistent_property_value("data_item_references", {k: v for k, v in data_item_references.items()})

    def clear_data_item_reference(self, key: str) -> None:
        data_item_references = self.data_item_references
        del data_item_references[key]
        self._set_persistent_property_value("data_item_references", {k: v for k, v in data_item_references.items()})

    def prune(self) -> None:
        self.__storage_system.prune()

    def migrate_to_latest(self) -> None:
        self.__storage_system.migrate_to_latest()
        self.__storage_system.load_properties()
        self.update_storage_system()  # reload the properties
        self.prepare_read_project()
        self.read_project()

    def unmount(self) -> None:
        while len(self.data_groups) > 0:
            self.unload_item("data_groups", len(self.data_groups) - 1)
        while len(self.connections) > 0:
            self.unload_item("connections", len(self.connections) - 1)
        while len(self.computations) > 0:
            self.unload_item("computations", len(self.computations) - 1)
        while len(self.data_structures) > 0:
            self.unload_item("data_structures", len(self.data_structures) - 1)
        while len(self.display_items) > 0:
            self.unload_item("display_items", len(self.display_items) - 1)
        while len(self.data_items) > 0:
            self.unload_item("data_items", len(self.data_items) - 1)


def data_item_factory(lookup_id: typing.Callable[[str], str]) -> DataItem.DataItem:
    data_item_uuid = uuid.UUID(lookup_id("uuid"))
    # TODO: typing hack for default arg
    large_format = typing.cast(typing.Callable[..., typing.Any], lookup_id)("__large_format", False)
    return DataItem.DataItem(item_uuid=data_item_uuid, large_format=large_format)


def display_item_factory(lookup_id: typing.Callable[[str], str]) -> DisplayItem.DisplayItem:
    display_item_uuid = uuid.UUID(lookup_id("uuid"))
    return DisplayItem.DisplayItem(item_uuid=display_item_uuid)


def computation_factory(lookup_id: typing.Callable[[str], str]) -> Symbolic.Computation:
    return Symbolic.Computation()


def data_structure_factory(lookup_id: typing.Callable[[str], str]) -> DataStructure.DataStructure:
    return DataStructure.DataStructure()



"""Notification dialog.

Watches notification sources for new notifications.

The notification dialog is a global resource, shared between windows.

Notifications are constructed and posted via the Notification module.

Example:

from nion.swift.model import Notification
notification = Notification.Notification("notification", "\N{Earth Globe Americas} World Monitor", "Global Warming Happening Now", "Global warming is the unusually rapid increase in Earth's average surface temperature over the past century primarily due to the greenhouse gases released as people burn fossil fuels.")
Notification.notify(notification)

Notifications can be dismissed by the user or programmatically by calling `notification.dismiss()`.
"""

from __future__ import annotations

# standard libraries
import functools
import gettext
import typing

# third party libraries
# None

# local libraries
from nion.swift.model import Notification
from nion.ui import CanvasItem
from nion.ui import Declarative
from nion.utils import Geometry
from nion.utils import Model
from nion.utils import ListModel
from nion.utils import Process
from nion.utils import Registry

if typing.TYPE_CHECKING:
    from nion.ui import Application
    from nion.ui import UserInterface
    from nion.ui import Window
    from nion.utils import Event

_ = gettext.gettext


class CharButtonConstructor:
    def construct(self, d_type: str, ui: UserInterface.UserInterface, window: typing.Optional[Window.Window],
                  d: Declarative.UIDescription, handler: Declarative.HandlerLike,
                  finishes: typing.List[typing.Callable[[], None]]) -> typing.Optional[UserInterface.Widget]:
        if d_type == "notification_char_button":
            text = d["text"]
            canvas_item = CanvasItem.TextButtonCanvasItem(text)
            canvas_item.text_font = "normal 13px serif"
            canvas_item.padding = Geometry.IntSize(canvas_item.padding.height, 0)
            canvas_item.size_to_content(ui.get_font_metrics)
            widget = ui.create_canvas_widget(properties={"height": canvas_item.sizing.preferred_height, "width": canvas_item.sizing.preferred_width})
            widget.canvas_item.add_canvas_item(canvas_item)
            if handler:
                Declarative.connect_name(widget, d, handler)
                Declarative.connect_attributes(widget, d, handler, finishes)
                Declarative.connect_event(widget, canvas_item, d, handler, "on_clicked", [])
            return widget
        return None


Registry.register_component(CharButtonConstructor(), {"declarative_constructor"})


class NotificationHandler(Declarative.Handler):
    """Declarative component handler for a section in a multiple acquire method component."""

    def __init__(self, notification: Notification.Notification) -> None:
        super().__init__()
        self.notification = notification
        u = Declarative.DeclarativeUI()
        self.ui_view = u.create_row(
            u.create_column(
                u.create_row(
                    u.create_label(text="@binding(notification.task_name)", color="#3366CC"),
                    u.create_stretch(),
                    {"type": "notification_char_button", "text": " \N{MULTIPLICATION X} ", "on_clicked": "handle_dismiss"},
                    spacing=8
                ),
                u.create_row(
                    u.create_label(text="@binding(notification.title)", font="bold"),
                    u.create_stretch(),
                    spacing=8
                ),
                u.create_row(
                    u.create_label(text="@binding(notification.text)", word_wrap=True, width=440),
                    u.create_stretch(),
                    spacing=8
                ),
                u.create_divider(orientation="horizontal"),
                spacing=4,
            ),
        )

    def handle_dismiss(self, widget: Declarative.UIWidget) -> None:
        self.notification.dismiss()



class NotificationComponentFactory(typing.Protocol):
    def make_component(self, app: Application.BaseApplication, notification: Notification.Notification) -> typing.Optional[Declarative.HandlerLike]: ...


class NotificationDialog(Declarative.WindowHandler):

    def __init__(self, app: Application.BaseApplication) -> None:
        super().__init__()

        self.app = app

        notifications = ListModel.ListModel[Notification.Notification]()
        stack_index = Model.PropertyModel[int](0)

        self.notifications = notifications
        self.stack_index = stack_index

        def notification_changed(key: str, value: Notification.Notification, index: int) -> None:
            stack_index.value = 1 if len(notifications.items) > 0 else 0
            if len(notifications.items) == 0:
                self.window.request_close()

        self.__notification_inserted_listener = self.notifications.item_inserted_event.listen(notification_changed)
        self.__notification_removed_listener = self.notifications.item_removed_event.listen(notification_changed)

        u = Declarative.DeclarativeUI()
        main_column = u.create_column(
            u.create_stack(
                u.create_column(
                    u.create_label(text=_("No notifications.")),
                    u.create_stretch(),
                ),
                u.create_column(
                    u.create_column(items="notifications.items", item_component_id="notification", spacing=6),
                    u.create_stretch(),
                ),
                current_index="@binding(stack_index.value)"
            ),
            u.create_stretch(),
            spacing=8,
            width=460,
            min_height=260,
        )

        window = u.create_window(main_column, title=_("Notifications"), margin=8, window_style="tool")
        self.run(window, app=app, persistent_id="notification_dialog")

    def close(self) -> None:
        global notification_dialog
        notification_dialog = None
        super().close()

    def create_handler(self, component_id: str, container: typing.Any = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
        # this is called to construct contained declarative component handlers within this handler.
        if component_id == "notification":
            assert container is not None
            assert item is not None
            notification = typing.cast(Notification.Notification, item)
            for component in Registry.get_components_by_type("notification-component-factory"):
                notification_handler = typing.cast(NotificationComponentFactory, component)
                notification_component = notification_handler.make_component(self.app, notification)
                if notification_component:
                    return notification_component
            return NotificationHandler(notification)
        return None


notification_dialog: typing.Optional[NotificationDialog] = None

_notification_sources: typing.List[Notification.NotificationSourceLike] = list()
_notification_source_listeners: typing.List[Event.EventListener] = list()
_notification_dismiss_listeners: typing.List[Event.EventListener] = list()


def notification_dismissed(notification: Notification.Notification) -> None:
    notifications = notification_dialog.notifications if notification_dialog else None
    if notifications:
        index = list(notifications.items).index(notification)
        _notification_dismiss_listeners.pop(index).close()
        notifications.remove_item(index)


def append_notification(notification: Notification.Notification) -> None:
    async def append_notification_async(notification: Notification.Notification) -> None:
        with Process.audit("append_notification"):
            open_notification_dialog()
            _notification_dismiss_listeners.append(notification.dismiss_event.listen(functools.partial(notification_dismissed, notification)))
            notifications = notification_dialog.notifications if notification_dialog else None
            if notifications:
                notifications.append_item(notification)

    if _app and _app.event_loop:
        _app.event_loop.create_task(append_notification_async(notification))


def component_registered(component: Registry._ComponentType, component_types: typing.Set[str]) -> None:
    if "notification-source" in component_types:
        notification_source = typing.cast(Notification.NotificationSourceLike, component)
        _notification_source_listeners.append(notification_source.notify_event.listen(append_notification))
        _notification_sources.append(notification_source)
        for notification in notification_source.notifications:
            append_notification(notification)


def component_unregistered(component: Registry._ComponentType, component_types: typing.Set[str]) -> None:
    if "notification-source" in component_types:
        notification_source = typing.cast(Notification.NotificationSourceLike, component)
        index = _notification_sources.index(notification_source)
        _notification_sources.pop(index)
        _notification_source_listeners.pop(index).close()


_component_registered_listener = Registry.listen_component_registered_event(component_registered)
_component_unregistered_listener = Registry.listen_component_unregistered_event(component_unregistered)

Registry.fire_existing_component_registered_events("notification-source")

_app: Application.BaseApplication = typing.cast(typing.Any, None)


def open_notification_dialog() -> None:
    global notification_dialog
    if not notification_dialog:
        notification_dialog = NotificationDialog(_app)


def close_notification_dialog() -> None:
    global notification_dialog
    if notification_dialog:
        notification_dialog.close()



# standard libraries
import datetime
import typing

# third party libraries
# None

# local libraries
import uuid

from nion.swift.model import Persistence
from nion.swift.model.Persistence import PersistentDictType
from nion.utils import Converter
from nion.utils import DateTime


class WorkspaceLayout(Persistence.PersistentObject):
    """
        Represents a specific layout available in the workspace.

        A layout consists of a set of panels within other canvas items and includes
        content of each of those panels.
    """
    def __init__(self) -> None:
        super().__init__()
        self.define_type("workspace")
        self.define_property("created", None, hidden=True, converter=Converter.DatetimeToStringConverter())
        self.define_property("name", str(), hidden=True)
        self.define_property("layout", dict(), hidden=True)
        self.define_property("workspace_id", str(uuid.uuid4()), hidden=True)

    def read_from_dict(self, properties: PersistentDictType) -> None:
        super().read_from_dict(properties)
        if self.created is None:  # invalid timestamp -- set property to modified but don't trigger change
            self._get_persistent_property("created").value = self.modified

    @property
    def created(self) -> datetime.datetime:
        return typing.cast(datetime.datetime, self._get_persistent_property_value("created"))

    @property
    def timestamp_for_sorting(self) -> datetime.datetime:
        created = self.created
        return created if created else self.modified

    @property
    def name(self) -> str:
        return typing.cast(str, self._get_persistent_property_value("name"))

    @name.setter
    def name(self, value: str) -> None:
        self._set_persistent_property_value("name", value)

    @property
    def layout(self) -> Persistence.PersistentDictType:
        return typing.cast(Persistence.PersistentDictType, self._get_persistent_property_value("layout"))

    @layout.setter
    def layout(self, value: Persistence.PersistentDictType) -> None:
        self._set_persistent_property_value("layout", value)

    @property
    def workspace_id(self) -> str:
        return typing.cast(str, self._get_persistent_property_value("workspace_id"))

    @workspace_id.setter
    def workspace_id(self, value: str) -> None:
        self._set_persistent_property_value("workspace_id", value)


def factory(lookup_id: typing.Callable[[str], str]) -> WorkspaceLayout:
    return WorkspaceLayout()
